@@ -25,9 +25,9 @@
 # are not necessary
 """Base module for trollflow2."""
 
-from pkg_resources import get_distribution, DistributionNotFound
-try:
-    __version__ = get_distribution(__name__).version
-except DistributionNotFound:
-    # package is not installed
-    pass
+from importlib.metadata import version
+from multiprocessing import Manager
+
+MP_MANAGER = Manager()
+
+__version__ = version(__name__)
