@@ -26,10 +26,11 @@ This delegate the actual running of the plugins to a subprocess to avoid any
 memory buildup.
 """
 
-
+import argparse
 import ast
 import copy
 import gc
+import logging
 import os
 import re
 import signal
@@ -37,25 +38,26 @@ import traceback
 from collections import OrderedDict
 from contextlib import suppress
 from datetime import datetime
-from logging import getLogger
 from queue import Empty
-from urllib.parse import urlparse
+from urllib.parse import urlsplit
 
 import yaml
-from trollflow2.dict_tools import gen_dict_extract, plist_iter
-from trollflow2.logging import setup_queued_logging
-from trollflow2.plugins import AbortProcessing
+from yaml import BaseLoader, SafeLoader, UnsafeLoader
 
 try:
     from posttroll.listener import ListenerContainer
 except ImportError:
     ListenerContainer = None
 
-from yaml import UnsafeLoader, SafeLoader, BaseLoader
+from trollflow2 import MP_MANAGER
+from trollflow2.dict_tools import gen_dict_extract, plist_iter
+from trollflow2.logging import (create_logged_process, logging_on,
+                                queued_logging)
+from trollflow2.plugins import AbortProcessing
 
-
-LOG = getLogger(__name__)
+logger = logging.getLogger(__name__)
 DEFAULT_PRIORITY = 999
+VALID_MESSAGE_TYPES = ("file", "dataset", "collection")
 
 
 def tuple_constructor(loader, node):
@@ -85,7 +87,7 @@ def get_test_message(test_message_file):
     return msg
 
 
-def check_results(produced_files, start_time, exitcode):
+def check_results(produced_files, start_time: datetime, exitcode: int):
     """Make sure the composites have been saved."""
     end_time = datetime.now()
     error_detected = False
@@ -97,42 +99,77 @@ def check_results(produced_files, start_time, exitcode):
         try:
             saved_file = produced_files.get(block=False)
             try:
-                if os.path.getsize(saved_file) == 0:
-                    LOG.error("Empty file detected: %s", saved_file)
-                    error_detected = True
+                error_detected = _check_file(saved_file)
             except FileNotFoundError:
-                LOG.error("Missing file: %s", saved_file)
+                logger.error("Missing file: %s", saved_file)
                 error_detected = True
+            except NotImplementedError as err:
+                logger.error(err)
+                error_detected = True
+            if error_detected:
+                break
         except Empty:
             break
     if exitcode != 0:
         error_detected = True
         if exitcode < 0:
-            LOG.error('Process killed with signal %d', -exitcode)
+            logger.error('Process killed with signal %d', -exitcode)
         else:
-            LOG.critical('Process crashed with exit code %d', exitcode)
+            logger.critical('Process crashed with exit code %d', exitcode)
     if not error_detected:
         elapsed = end_time - start_time
         if qsize is not None:
-            LOG.info(f'All {qsize:d} files produced nominally in '
-                     f"{elapsed!s}", extra={"time": elapsed})
+            logger.info(f'All {qsize:d} files produced nominally in '
+                        f"{elapsed!s}", extra={"time": elapsed})
         else:
-            LOG.info(f'All files produced nominally in '
-                     f"{elapsed!s}", extra={"time": elapsed})
+            logger.info(f'All files produced nominally in '
+                        f"{elapsed!s}", extra={"time": elapsed})
+
+
+def _check_file(saved_file):
+    file_scheme = urlsplit(saved_file).scheme
+    if file_scheme in ('', 'file'):
+        return _check_local_file(saved_file)
+    if file_scheme == 's3':
+        from trollflow2.plugins.s3 import check_s3_file
+        return check_s3_file(saved_file)
+    raise NotImplementedError("File check not impleneted for remote filesystem %s" % file_scheme)
+
+
+def _check_local_file(saved_file):
+    if os.path.getsize(saved_file) == 0:
+        logger.error("Empty file detected: %s", saved_file)
+        return True
+    return False
 
 
 def generate_messages(connection_parameters):
     """Generate messages using a ListenerContainer."""
     listener = _create_listener_from_connection_parameters(connection_parameters)
-    while True:
+    keep_looping = True
+
+    def _signal_handler(signum, frame):
+        nonlocal keep_looping
+        logger.info(
+            "Caught signal to stop processing new messages and to terminate Trollflow2.")
+        keep_looping = False
+
+    signal.signal(signal.SIGTERM, _signal_handler)
+
+    while keep_looping:
         try:
-            yield listener.output_queue.get(True, 5)
+            msg = listener.output_queue.get(True, 1)
+            if msg.type in VALID_MESSAGE_TYPES:
+                logger.info("New message received.")
+                logger.debug(f"{str(msg)}")
+                yield msg
         except KeyboardInterrupt:
-            listener.stop()
-            return
+            break
         except Empty:
             continue
 
+    listener.stop()
+
 
 def _create_listener_from_connection_parameters(connection_parameters):
     """Create listener from connection parameters."""
@@ -149,10 +186,10 @@ def _create_listener_from_connection_parameters(connection_parameters):
 class Runner:
     """Class that handles all the administration around running on a product list."""
 
-    def __init__(self, product_list, log_queue, connection_parameters=None, test_message=None, threaded=False):
+    def __init__(self, product_list, connection_parameters=None,
+                 test_message=None, threaded=False):
         """Set up the runner."""
         self.product_list = product_list
-        self.log_queue = log_queue
         self.connection_parameters = connection_parameters
         self.test_message = get_test_message(test_message)
         self.threaded = threaded
@@ -188,25 +225,21 @@ class Runner:
 
     def _run_threaded(self, messages):
         """Run in a thread."""
-        LOG.info("Launching trollflow2 with threads")
+        logger.debug("Launching trollflow2 with threads")
         from threading import Thread
         self._run_product_list_on_messages(messages, process, Thread)
 
     def _run_subprocess(self, messages):
         """Run in a subprocess, with queued logging."""
-        LOG.info("Launching trollflow2 with subprocesses")
-        from multiprocessing import get_context
-        ctx = get_context("spawn")
-        self._run_product_list_on_messages(messages, queue_logged_process, ctx.Process)
+        logger.debug("Launching trollflow2 with subprocesses")
+        self._run_product_list_on_messages(messages, queue_logged_process, create_logged_process)
 
-    def _run_product_list_on_messages(self, messages, target_fun, process_class):
+    def _run_product_list_on_messages(self, messages, target_fun, process_creator):
         """Run the product list on the messages."""
         for msg in messages:
-            produced_files_queue = self.log_queue._manager.Queue()
+            produced_files_queue = MP_MANAGER.Queue()
             kwargs = dict(produced_files=produced_files_queue, prod_list=self.product_list)
-            if not self.threaded:
-                kwargs["log_queue"] = self.log_queue
-            proc = process_class(target=target_fun, args=(msg,), kwargs=kwargs)
+            proc = process_creator(target=target_fun, args=(msg,), kwargs=kwargs)
             start_time = datetime.now()
             proc.start()
             proc.join()
@@ -235,6 +268,13 @@ def get_area_priorities(product_list):
 
 def message_to_jobs(msg, product_list):
     """Convert a posttroll message *msg* to a list of jobs given a *product_list*."""
+    input_filenames = _extract_filenames(msg)
+    input_mda = msg.data
+    return file_list_to_jobs(input_filenames, product_list, input_mda)
+
+
+def file_list_to_jobs(input_filenames, product_list, input_mda):
+    """Convert a file list to jobs."""
     formats = product_list['product_list'].get('formats', None)
     for _product, pconfig in plist_iter(product_list['product_list'], level='product'):
         if 'formats' not in pconfig and formats is not None:
@@ -242,11 +282,10 @@ def message_to_jobs(msg, product_list):
     jobs = OrderedDict()
     priorities = get_area_priorities(product_list)
     # TODO: check the uri is accessible from the current host.
-    input_filenames = _extract_filenames(msg)
     for prio, areas in priorities.items():
         jobs[prio] = OrderedDict()
         jobs[prio]['input_filenames'] = input_filenames.copy()
-        jobs[prio]['input_mda'] = msg.data.copy()
+        jobs[prio]['input_mda'] = input_mda.copy()
         jobs[prio]['product_list'] = {}
         for section in product_list:
             if section == 'product_list':
@@ -266,21 +305,27 @@ def _extract_filenames(msg):
 
     If the message contains a `filesystem` item, use fsspec to decode it.
     """
-    filenames = [urlparse(uri).path for uri in gen_dict_extract(msg.data, 'uri')]
-    filenames = _create_fs_file_instances(filenames, msg)
-    return filenames
-
-
-def _create_fs_file_instances(filenames, msg):
-    """Create FSFile instances when filesystem is provided."""
     filesystems = list(gen_dict_extract(msg.data, 'filesystem'))
     if filesystems:
-        from satpy.readers import FSFile
-        from fsspec.spec import AbstractFileSystem
-        import json
-        filenames = [FSFile(filename, AbstractFileSystem.from_json(json.dumps(filesystem)))
-                     for filename, filesystem in zip(filenames, filesystems)]
-    return filenames
+        filenames = [filename for filename in gen_dict_extract(msg.data, 'path')]
+        if not filenames:
+            filenames = [filename for filename in gen_dict_extract(msg.data, 'uid')]
+        to_open = _create_fs_files(filenames, filesystems)
+    else:
+        to_open = [uri for uri in gen_dict_extract(msg.data, 'uri')]
+
+    return to_open
+
+
+def _create_fs_files(filenames, filesystems):
+    """Create FSFile instances when filesystem is provided."""
+    import json
+
+    from fsspec.spec import AbstractFileSystem
+    from satpy.readers import FSFile
+    fsfiles = [FSFile(filename, AbstractFileSystem.from_json(json.dumps(filesystem)))
+               for filename, filesystem in zip(filenames, filesystems)]
+    return fsfiles
 
 
 def expand(yml):
@@ -296,7 +341,7 @@ def expand(yml):
     return yml
 
 
-def get_dask_client(config):
+def get_dask_distributed_client(config):
     """Create Dask client if configured."""
     client = None
 
@@ -306,29 +351,32 @@ def get_dask_client(config):
         client = client_class(**settings)
         try:
             if not client.ncores():
-                LOG.warning("No workers available, reverting to default scheduler")
+                logger.warning("No workers available, reverting to default scheduler")
                 client.close()
                 client = None
         except AttributeError:
             client = None
     except OSError:
-        LOG.error("Scheduler not found, reverting to default scheduler")
+        logger.error("Scheduler not found, reverting to default scheduler")
     except KeyError:
-        LOG.debug("Distributed processing not configured, "
-                  "using default scheduler")
+        logger.debug("Distributed processing not configured, "
+                     "using default scheduler")
     else:
-        LOG.debug(f"Using dask distributed client {client!s}")
+        logger.debug(f"Using dask distributed client {client!s}")
 
     return client
 
 
-def queue_logged_process(msg, prod_list, produced_files, log_queue):
+@queued_logging
+def queue_logged_process(msg, prod_list, produced_files):
     """Run `process` with a queued log."""
-    setup_queued_logging(log_queue)
     with suppress(ValueError):
         signal.signal(signal.SIGUSR1, print_traces)
-        LOG.debug("Use SIGUSR1 on pid {} to check the current tracebacks of this subprocess.".format(os.getpid()))
-    process(msg, prod_list, produced_files)
+        logger.debug("Use SIGUSR1 on pid {} to check the current tracebacks of this subprocess.".format(os.getpid()))
+    try:
+        process(msg, prod_list, produced_files)
+    finally:
+        logging.shutdown()
 
 
 def print_traces(signum, frame):
@@ -345,40 +393,22 @@ def print_traces(signum, frame):
 
 def process(msg, prod_list, produced_files):
     """Process a message."""
+    """Convert a posttroll message *msg* to a list of jobs given a *product_list*."""
+    input_filenames = _extract_filenames(msg)
+    input_mda = msg.data
+    process_files(input_filenames, input_mda, prod_list, produced_files)
+
+
+def process_files(input_filenames, input_mda, prod_list, produced_files):
+    """Process files."""
     config = read_config(prod_list, Loader=UnsafeLoader)
-
-    # Get distributed client
-    client = get_dask_client(config)
-
+    client = get_dask_distributed_client(config)
     try:
         config = expand(config)
-        jobs = message_to_jobs(msg, config)
-        for prio in sorted(jobs.keys()):
-            job = jobs[prio]
-            job['processing_priority'] = prio
-            job['produced_files'] = produced_files
-            try:
-                for wrk in config['workers']:
-                    cwrk = wrk.copy()
-                    if "timeout" in cwrk:
-
-                        def _timeout_handler(signum, frame, wrk=wrk):
-                            raise TimeoutError(
-                                f"Timeout for {wrk['fun']!s} expired "
-                                f"after {wrk['timeout']:.1f} seconds, "
-                                "giving up")
-                        signal.signal(signal.SIGALRM, _timeout_handler)
-                        # using setitimer because it accepts floats,
-                        # unlike signal.alarm
-                        signal.setitimer(signal.ITIMER_REAL,
-                                         cwrk.pop("timeout"))
-                    cwrk.pop('fun')(job, **cwrk)
-                    if "timeout" in cwrk:
-                        signal.alarm(0)  # cancel the alarm
-            except AbortProcessing as err:
-                LOG.warning(str(err))
+        jobs = file_list_to_jobs(input_filenames, config, input_mda)
+        process_jobs(config["workers"], jobs, produced_files)
     except Exception:
-        LOG.exception("Process crashed")
+        logger.exception("Process crashed")
         if "crash_handlers" in config:
             trace = traceback.format_exc()
             for hand in config['crash_handlers']['handlers']:
@@ -387,32 +417,61 @@ def process(msg, prod_list, produced_files):
     finally:
         # Remove config and run garbage collection so all remaining
         # references e.g. to FilePublisher should be removed
-        LOG.debug('Cleaning up')
+        logger.debug('Cleaning up')
         for wrk in config.get("workers", []):
             try:
                 wrk['fun'].stop()
             except AttributeError:
                 continue
         del config
-        try:
+        with suppress(AttributeError):
             client.close()
-        except AttributeError:
-            pass
         gc.collect()
 
 
+def process_jobs(workers, jobs, produced_files):
+    """Process the jobs."""
+    for prio in sorted(jobs.keys()):
+        job = jobs[prio]
+        job['processing_priority'] = prio
+        job['produced_files'] = produced_files
+        try:
+            for wrk in workers:
+                cwrk = wrk.copy()
+                if "timeout" in cwrk:
+                    def _timeout_handler(signum, frame, wrk=wrk):
+                        raise TimeoutError(
+                            f"Timeout for {wrk['fun']!s} expired "
+                            f"after {wrk['timeout']:.1f} seconds, "
+                            "giving up")
+
+                    signal.signal(signal.SIGALRM, _timeout_handler)
+                    # using setitimer because it accepts floats,
+                    # unlike signal.alarm
+                    signal.setitimer(signal.ITIMER_REAL,
+                                     cwrk.pop("timeout"))
+                cwrk.pop('fun')(job, **cwrk)
+                if "timeout" in cwrk:
+                    signal.alarm(0)  # cancel the alarm
+        except AbortProcessing as err:
+            logger.warning(str(err))
+
+
 def read_config(fname=None, raw_string=None, Loader=SafeLoader):
     """Read the configuration file."""
     try:
         if fname:
+            logger.debug(f"Reading config file {fname}")
             with open(fname) as fid:
                 raw_config = fid.read()
         elif raw_string:
             raw_config = raw_string
+        if not raw_config:
+            raise IOError
         config = yaml.load(_remove_null_keys(raw_config), Loader=Loader)
     except (IOError, yaml.YAMLError):
         # Either open() or yaml.load() failed
-        LOG.exception("Process crashed, check YAML file.")
+        logger.exception("Process crashed, check YAML file.")
         raise
     return config
 
@@ -424,7 +483,7 @@ def _remove_null_keys(raw_config):
 def sendmail(config, trace):
     """Send email about crashes using `sendmail`."""
     from email.mime.text import MIMEText
-    from subprocess import Popen, PIPE
+    from subprocess import PIPE, Popen
 
     email_settings = config['sendmail']
     msg = MIMEText(email_settings["header"] + "\n\n" + "\n\n" + trace)
@@ -436,3 +495,62 @@ def sendmail(config, trace):
     pid = Popen([sendmail, "-t", "-oi"], stdin=PIPE)
     pid.communicate(msg.as_bytes())
     pid.terminate()
+
+
+def launch(args_in):
+    """Launch the processing."""
+    args = parse_args(args_in)
+
+    log_config = _read_log_config(args)
+
+    with logging_on(log_config):
+        logger.info("Launching Satpy-based runner.")
+        product_list = args.pop("product_list")
+        test_message = args.pop("test_message")
+        threaded = args.pop("threaded")
+        connection_parameters = args
+
+        runner = Runner(product_list, connection_parameters, test_message, threaded)
+        runner.run()
+
+
+def _read_log_config(args):
+    log_config = args.pop("log_config", None)
+    if log_config is not None:
+        with open(log_config) as fd:
+            log_config = yaml.safe_load(fd.read())
+    return log_config
+
+
+def parse_args(args_in):
+    """Parse commandline arguments."""
+    parser = argparse.ArgumentParser(
+        description='Launch trollflow2 processing with Satpy listening on the specified Posttroll topic(s)')
+    parser.add_argument("topic", nargs='*',
+                        help="Topic to listen to",
+                        type=str)
+    parser.add_argument("product_list",
+                        help="The yaml file with the product list",
+                        type=str)
+    parser.add_argument("-m", "--test_message",
+                        help="File path with the message used for testing offline. This implies threaded running.",
+                        type=str, required=False)
+    parser.add_argument("-t", "--threaded",
+                        help="Run the product generation in threads instead of processes.",
+                        action='store_true')
+
+    parser.add_argument("-c", "--log-config",
+                        help="Log config file (yaml) to use",
+                        type=str, required=False)
+    parser.add_argument('-n', "--nameserver", required=False, type=str,
+                        help="Nameserver to connect to. Disable by setting to False", default='localhost')
+    parser.add_argument('-a', "--addresses", required=False, type=str,
+                        help=("Add direct TCP port connection.  Can be used several times: "
+                              "'-a tcp://127.0.0.1:12345 -a tcp://123.456.789.0:9013'"),
+                        action="append")
+
+    args = vars(parser.parse_args(args_in))
+    if args['nameserver'].lower() in ('false', 'off', '0'):
+        args['nameserver'] = False
+
+    return args
